/* filelist.c - for drawing a scrollable filelist with size, permissions etc.
   Copyright (C) 1997 Paul Sheer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <my_string.h>
#include "stringtools.h"
#include <sys/types.h>

#if HAVE_DIRENT_H
#include <dirent.h>
#define NAMLEN(dirent) strlen((dirent)->d_name)
#else
#define dirent direct
#define NAMLEN(dirent) (dirent)->d_namlen
#if HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#if HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#if HAVE_NDIR_H
#include <ndir.h>
#endif
#endif

#include <sys/stat.h>
#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "coolwidget.h"
#include "pool.h"
#include "mad.h"

/*
 * struct file_entry {
 *     unsigned long options;
 *     struct dirent dirent;
 *     struct stat stat;
 * };
 */

/* This doesn't apply to solaris: */
/*
 * struct dirent {
 *      long            d_ino;
 *      __kernel_off_t  d_off;
 *      unsigned short  d_reclen;
 *      char            d_name[256];
 * };
 */

/*

 * struct new_stat {
 *      unsigned short st_dev;
 *      unsigned short __pad1;
 *      unsigned long st_ino;
 *      unsigned short st_mode;
 *      unsigned short st_nlink;
 *      unsigned short st_uid;
 *      unsigned short st_gid;
 *      unsigned short st_rdev;
 *      unsigned short __pad2;
 *      unsigned long  st_size;
 *      unsigned long  st_blksize;
 *      unsigned long  st_blocks;
 *      unsigned long  st_atime;
 *      unsigned long  __unused1;
 *      unsigned long  st_mtime;
 *      unsigned long  __unused2;
 *      unsigned long  st_ctime;
 *      unsigned long  __unused3;
 *      unsigned long  __unused4;
 *      unsigned long  __unused5;
 * };
 * 
 */

/*

 * #define S_ISLNK(m)   (((m) & S_IFMT) == S_IFLNK)
 * #define S_ISREG(m)   (((m) & S_IFMT) == S_IFREG)
 * #define S_ISDIR(m)   (((m) & S_IFMT) == S_IFDIR)
 * #define S_ISCHR(m)   (((m) & S_IFMT) == S_IFCHR)
 * #define S_ISBLK(m)   (((m) & S_IFMT) == S_IFBLK)
 * #define S_ISFIFO(m)  (((m) & S_IFMT) == S_IFIFO)
 * #define S_ISSOCK(m)  (((m) & S_IFMT) == S_IFSOCK)
 * 
 * #define S_IRWXU 00700
 * #define S_IRUSR 00400
 * #define S_IWUSR 00200
 * #define S_IXUSR 00100
 * 
 * #define S_IRWXG 00070
 * #define S_IRGRP 00040
 * #define S_IWGRP 00020
 * #define S_IXGRP 00010
 * 
 * #define S_IRWXO 00007
 * #define S_IROTH 00004
 * #define S_IWOTH 00002
 * #define S_IXOTH 00001
 */


char *dname (struct dirent *directentry);

#ifdef HAVE_STRFTIME
/* We want our own dates for NLS */
#undef HAVE_STRFTIME
#endif

#undef gettext_noop
#define gettext_noop(x) x

/* reset with timestr = 0 */
void get_file_time (char *timestr, time_t file_time)
{
#ifndef HAVE_STRFTIME
    static char monthstr[12][8] =
    {
/* Month list */
	gettext_noop ("Jan"), gettext_noop ("Feb"), gettext_noop ("Mar"), gettext_noop ("Apr"), gettext_noop ("May"), gettext_noop ("Jun"), gettext_noop ("Jul"), gettext_noop ("Aug"), gettext_noop ("Sep"), gettext_noop ("Oct"), gettext_noop ("Nov"), gettext_noop ("Dec")
    };
#endif
    static struct tm tm_current =
    {-1};
    struct tm *tm;
    static int i = 0;

    if (!timestr) {
	i = 0;
	return;
    }
    if (!i)
	for (i = 0; i < 12; i++)
	    strcpy (monthstr[i], _ (monthstr[i]));

    if (tm_current.tm_sec == -1) {
	time_t t;
	time (&t);
	memcpy ((void *) &tm_current, (void *) localtime (&t), sizeof (struct tm));
    }
    tm = localtime (&(file_time));
#if HAVE_STRFTIME
    if (tm->tm_year == tm_current.tm_year)	/* date with year and without time */
	strftime (timestr, 31, "%b %d %H:%M", tm);
    else			/* date without year and with time */
	strftime (timestr, 31, "%Y %b %d", tm);
#else
    if (tm->tm_year == tm_current.tm_year)	/* date with year and without time */
	sprintf (timestr, "%s %.2d %.2d:%.2d", monthstr[tm->tm_mon],
		 tm->tm_mday, tm->tm_hour, tm->tm_min);
    else			/* date without year and with time */
	sprintf (timestr, "%d %s %.2d", tm->tm_year + 1900,
		 monthstr[tm->tm_mon], tm->tm_mday);
#endif
}


char **get_filelist_line (void *data, int line_number, int *num_fields, int *tagged)
{
    struct file_entry *directentry;
    static char *fields[10], size[24], mode[12], timestr[32];
    static char name[520], *n;
    mode_t m;

    *num_fields = 4;		/* name, size, date, mode only (for the mean time) */

    directentry = (struct file_entry *) data;
    if (directentry[line_number].options & FILELIST_LAST_ENTRY)
	return 0;

    n = name;
    strcpy (name, directentry[line_number].name);
    fields[0] = name;
    sprintf (size, "\t%u", (unsigned int) directentry[line_number].stat.st_size);
    fields[1] = size;

    get_file_time (timestr, directentry[line_number].stat.st_mtime);
    fields[2] = timestr;

    memset (mode, ' ', 11);
    mode[11] = 0;
    mode[0] = '-';
    m = directentry[line_number].stat.st_mode;
    switch ((int) m & S_IFMT) {
    case S_IFLNK:
	mode[0] = 'l';
	break;
    case S_IFDIR:
	mode[0] = 'd';
	break;
    case S_IFCHR:
	mode[0] = 'c';
	break;
    case S_IFBLK:
	mode[0] = 'b';
	break;
    case S_IFIFO:
	mode[0] = 'f';
	break;
    case S_IFSOCK:
	mode[0] = 's';
	break;
    }

    mode[1] = m & S_IRUSR ? 'r' : '-';
    mode[2] = m & S_IWUSR ? 'w' : '-';
    mode[3] = m & S_IXUSR ? 'x' : '-';

    mode[4] = m & S_IRGRP ? 'r' : '-';
    mode[5] = m & S_IWGRP ? 'w' : '-';
    mode[6] = m & S_IXGRP ? 'x' : '-';

    mode[7] = m & S_IROTH ? 'r' : '-';
    mode[8] = m & S_IWOTH ? 'w' : '-';
    mode[9] = m & S_IXOTH ? 'x' : '-';

    if (S_ISLNK (m)) {
	int l, i;
	char *p;
	p = directentry[line_number].name;
	l = strlen (n);
	for (i = 0; i < l; i++) {
	    *n++ = '\b';
	    *n++ = *p++;
	}
	*n++ = '\0';
    } else if (m & (S_IXUSR | S_IXGRP | S_IXOTH)) {
	int l, i;
	char *p;
	p = directentry[line_number].name;
	l = strlen (n);
	for (i = 0; i < l; i++) {
	    *n++ = '\r';
	    *n++ = *p++;
	}
	*n++ = '\0';
    }
    fields[3] = mode;
    fields[*num_fields] = 0;
    if (directentry[line_number].options & FILELIST_TAGGED_ENTRY)
	*tagged = 1;
    return fields;
}

char *get_filelist_line_short (void *data, int line_number, char buffer[512])
{
    struct file_entry *directentry;
    static char ctimestr[32], mtimestr[32];
    mode_t m;
    char *ptr;
    unsigned long size_bytes;

    directentry = (struct file_entry *) data;
    if (directentry[line_number].options & FILELIST_LAST_ENTRY)
	return 0;

    buffer[0] = 6;
    strncpy (&(buffer[1]), directentry[line_number].name, 256);
    ptr = &(buffer[strlen (buffer)]);

    sprintf (ptr, "%c\nMode: %c", 27, 9);
    ptr += strlen (ptr);

    m = directentry[line_number].stat.st_mode;
    switch ((int) m & S_IFMT) {
    case S_IFLNK:
	*ptr = 'l';
	break;
    case S_IFDIR:
	*ptr = 'd';
	break;
    case S_IFCHR:
	*ptr = 'c';
	break;
    case S_IFBLK:
	*ptr = 'b';
	break;
    case S_IFIFO:
	*ptr = 'f';
	break;
    case S_IFSOCK:
	*ptr = 's';
	break;
    default:
	*ptr = '-';
    }
    ptr++;
    *(ptr++) = m & S_IRUSR ? 'r' : '-';
    *(ptr++) = m & S_IWUSR ? 'w' : '-';
    *(ptr++) = m & S_IXUSR ? 'x' : '-';

    *(ptr++) = m & S_IRGRP ? 'r' : '-';
    *(ptr++) = m & S_IWGRP ? 'w' : '-';
    *(ptr++) = m & S_IXGRP ? 'x' : '-';

    *(ptr++) = m & S_IROTH ? 'r' : '-';
    *(ptr++) = m & S_IWOTH ? 'w' : '-';
    *(ptr++) = m & S_IXOTH ? 'x' : '-';


    size_bytes = directentry[line_number].stat.st_size;
    sprintf (ptr, "%c ; Size: [%c%9lu%c] bytes or [%c%6lu%c] KBytes", 27, 7, size_bytes, 27, 7, size_bytes >> 10, 27);
    ptr += strlen (ptr);

    get_file_time (ctimestr, directentry[line_number].stat.st_ctime);
    get_file_time (mtimestr, directentry[line_number].stat.st_mtime);

    sprintf (ptr, "\nCreated on: %c%s%c ; Modified on: %c%s%c", 1, ctimestr, 27, 1, mtimestr, 27);

    return buffer;
}

Bool is_directory (struct file_entry * directentry, int line_number)
{
    if (directentry[line_number].options & FILELIST_LAST_ENTRY)
	return False;
    switch ((int) (directentry[line_number].stat.st_mode) & S_IFMT) {
    case S_IFLNK:
	/* fixme : add check if it is linked to dir or file */
	return True;
    case S_IFDIR:
	return True;
    }
    return False;
}

typedef struct FilelistCache {
    Pixmap cache;
    unsigned int width, height;
    unsigned int firstline, hilited;
    int current;
    unsigned int rowheight;
    unsigned long options;
} FilelistCache;

void free_filelist_cache (void *vcache)
{
    if (vcache) {
	FilelistCache *cache = (FilelistCache *) vcache;
	if (cache->cache) {
	    XFreePixmap (CDisplay, cache->cache);
	    cache->cache = None;
	}
	free (vcache);
    }
}

#define NEXT_ARROW_SIZE  	8
#define NEXT_ARROW_FIELD 	(NEXT_ARROW_SIZE*2)

void restrict_text_area (FilelistCache * cache)
{
    XRectangle clip_rec;
    clip_rec.x = 1;
    clip_rec.y = 1;
    clip_rec.width = cache->width - NEXT_ARROW_FIELD;
    clip_rec.height = cache->height - 2;
    XSetClipRectangles (CDisplay, CGC, 0, 0, &clip_rec, 1, YXSorted);
}

void push_filelist_line (FilelistCache * cache, Bool pushed)
{
    int y1, y2;
    if (cache->current < 0)
	return;
    y1 = (cache->current - cache->firstline) * cache->rowheight;
    y2 = y1 + cache->rowheight;

    if (y1 < cache->height && y2 > 0) {
	if (pushed)
	    CSetColor (COLOR_BLACK);
	else
	    CSetColor ((cache->hilited == cache->current) ? COLOR_WHITE : COLOR_FLAT);

	CLine (cache->cache, 0, y2, 0, y1);
	CLine (cache->cache, 1, y1, cache->width - 2, y1);

	if (pushed)
	    CSetColor ((cache->hilited == cache->current) ? COLOR_FLAT : COLOR_WHITE);

	CLine (cache->cache, 0, y2, cache->width - 2, y2);
	CLine (cache->cache, cache->width - 2, y2, cache->width - 2, y1);
    }
}

void hilite_filelist_line (FilelistCache * cache, int color, struct file_entry *directentry)
{
    int y1, y2, x2;
    y1 = (cache->hilited - cache->firstline) * cache->rowheight;
    y2 = y1 + (cache->rowheight / 2 - 4);
    x2 = cache->width - NEXT_ARROW_FIELD + NEXT_ARROW_SIZE / 2;

    if (y1 >= 0 && y1 < cache->height - FONT_OVERHEAD) {
	CSetColor (color);
	CRectangle (cache->cache, 0, y1, cache->width - 1, cache->rowheight - 1);
	CSetColor (COLOR_BLACK);
	if (!(directentry[cache->hilited].options & FILELIST_LAST_ENTRY) &&
	    directentry[cache->hilited].name != NULL) {
	    restrict_text_area (cache);
	    XDrawString (CDisplay, cache->cache, CGC,
			 FONT_OFFSET_X + 1, FONT_OFFSET_Y + y1,
			 directentry[cache->hilited].name,
			 strlen (directentry[cache->hilited].name));
	    XSetClipMask (CDisplay, CGC, None);
	    if (is_directory (directentry, cache->hilited)) {
		CSetColor (COLOR_BLACK);
		CLine (cache->cache, x2, y2, x2 + 7, y2 + 4);
		CLine (cache->cache, x2, y2, x2, y2 + 8);
		CSetColor (COLOR_FLAT);
		CLine (cache->cache, x2, y2 + 8, x2 + 7, y2 + 4);
	    }
	}
    }
}

void scroll_cache (FilelistCache * cache, unsigned int new_first, int *start, int *end)
{
    int good_start, good_end, scroll_to;

    if (cache->firstline > new_first) {		/* scrolling up */
	good_start = 0;
	scroll_to = (cache->firstline - new_first) * cache->rowheight;
	good_end = cache->height - scroll_to;
    } else {
	good_start = (new_first - cache->firstline) * cache->rowheight;
	good_end = cache->height;
	scroll_to = 0;
    }
    if (good_start < good_end) {
	XCopyArea (CDisplay, cache->cache, cache->cache, CGC, 0, good_start, cache->width, good_end, 0, scroll_to);
	if (cache->firstline > new_first) {	/* scrolling up */
	    *start = 0;
	    *end = scroll_to;
	} else {
	    *start = (cache->height - good_start);
	    *end = cache->height;
	}
    } else {
	*start = 0;
	*end = cache->height;
    }
    if (*start < *end) {
	CSetColor (COLOR_FLAT);
	CRectangle (cache->cache, 0, *start, cache->width, *end);
    }
    *start /= cache->rowheight;
    *end /= cache->rowheight;
}

void check_filelist_cache (CWidget * wdt, unsigned int w, unsigned int h, FilelistCache * cache)
{
    struct file_entry *directentry = (struct file_entry *) wdt->hook;
    int y, x, start_y;
    int i, start = 0, end = 0;

    if (cache->cache) {
	if (cache->width == w && cache->height == h) {
	    if (cache->firstline == wdt->firstline &&
		cache->hilited == wdt->cursor &&
		cache->options == wdt->options)
		return;
	} else {
	    XFreePixmap (CDisplay, cache->cache);
	    cache->cache = None;
	}
    }
    if (cache->cache == None) {
	cache->cache = XCreatePixmap (CDisplay, wdt->winid, w, h, CDepth);
	cache->width = w;
	cache->height = h;
	CTextSize (NULL, (int *) &(cache->rowheight), "Ty~g$V|&^.,/_'I}[@");
	if (cache->rowheight == 0)
	    cache->rowheight = 1;
	cache->rowheight += FONT_OVERHEAD;
	CSetColor (COLOR_FLAT);
	CRectangle (cache->cache, 0, 0, w, h);
	end = h / cache->rowheight;
    } else {
	push_filelist_line (cache, False);
	if (cache->hilited != wdt->cursor)
	    hilite_filelist_line (cache, COLOR_FLAT, directentry);

	if (cache->firstline != wdt->firstline)
	    scroll_cache (cache, wdt->firstline, &start, &end);
    }

    cache->firstline = wdt->firstline;
    cache->hilited = wdt->cursor;
    hilite_filelist_line (cache, (wdt->options & BUTTON_HIGHLIGHT) ? COLOR_WHITE : color_widget (12), directentry);
    cache->current = wdt->current;
    push_filelist_line (cache, True);

    cache->options = wdt->options;

    start_y = start * cache->rowheight;
    y = start_y + FONT_OFFSET_Y;
    start += wdt->firstline;
    end += wdt->firstline;

    CSetColor (COLOR_BLACK);
    restrict_text_area (cache);
    for (i = start;
	 i <= end &&
	 i < wdt->numlines && y < cache->height &&
	 !(directentry[i].options & FILELIST_LAST_ENTRY); i++) {
	if (directentry[i].name && i != cache->hilited)
	    XDrawString (CDisplay, cache->cache, CGC,
			 FONT_OFFSET_X + 1, y,
			 directentry[i].name,
			 strlen (directentry[i].name));

	y += cache->rowheight;
    }
    XSetClipMask (CDisplay, CGC, None);
    /* now drawing  arrows */
    y = start_y + (cache->rowheight / 2 - 4);
    x = cache->width - NEXT_ARROW_FIELD + NEXT_ARROW_SIZE / 2;
    for (i = start;
	 i <= end &&
	 i < wdt->numlines && y < cache->height &&
	 !(directentry[i].options & FILELIST_LAST_ENTRY); i++) {
	if (is_directory (directentry, i)) {
	    CSetColor (COLOR_BLACK);
	    CLine (cache->cache, x, y, x + 7, y + 4);
	    CLine (cache->cache, x, y, x, y + 8);
	    CSetColor ((i == cache->hilited) ? COLOR_FLAT : COLOR_WHITE);
	    CLine (cache->cache, x, y + 8, x + 7, y + 4);
	}
	y += cache->rowheight;
    }
}

void render_NeXT_filelist (CWidget * wdt)
{
    FilelistCache *cache;

    if (wdt->hook == NULL) {
	render_bevel (wdt->winid, 0, 0, wdt->width - 1, wdt->height - 1, 1, 3);
	return;
    }
    if (wdt->user == NULL) {
	wdt->user = calloc (sizeof (FilelistCache), 1);
	wdt->free_user = free_filelist_cache;
    }
    cache = (FilelistCache *) (wdt->user);

    CPushFont ("widget", 0);
    check_filelist_cache (wdt, wdt->width - 3, wdt->height - 3, cache);
    CPopFont ();

    render_bevel (wdt->winid, 0, 0, wdt->width - 1, wdt->height - 1, 1, 1);

    XCopyArea (CDisplay, cache->cache, wdt->winid, CGC, 0, 0, cache->width, cache->height, 2, 2);
}
void link_scrollbar_to_NeXT_filelist (CWidget * scrollbar, CWidget * w, XEvent * xevent, CEvent * cwevent, int whichscrbutton)
{
    /* fix me : add stuf */
    int redrawtext = 0, count, c;
    int new_first = w->firstline;
    static int r = 0;
    FilelistCache *cache;
    int box_size;

    if (w->user == NULL)
	return;
    cache = (FilelistCache *) (w->user);
    box_size = cache->height / cache->rowheight;

    if ((xevent->type == ButtonRelease || xevent->type == MotionNotify) && whichscrbutton == 3) {
	new_first = (double) scrollbar->firstline * w->numlines / 65535.0;
    } else if (xevent->type == ButtonPress && (cwevent->button == Button1 || cwevent->button == Button2)) {
	new_first = w->firstline;
	switch (whichscrbutton) {
	case 1:
	    new_first -= box_size - 1;
	    break;
	case 2:
	    new_first--;
	    break;
	case 5:
	    new_first++;
	    break;
	case 4:
	    new_first += box_size - 1;
	    break;
	}
    }
    if (new_first < 0)
	new_first = 0;
    else if (new_first > w->numlines - box_size)
	new_first = w->numlines - box_size;
    redrawtext = (new_first != w->firstline);
    w->firstline = new_first;

    if (xevent->type == ButtonRelease ||
	(redrawtext && !CCheckWindowEvent (xevent->xany.window, ButtonReleaseMask | ButtonMotionMask, 0)))
	render_NeXT_filelist (w);

    scrollbar->firstline = (double) 65535.0 *w->firstline / (w->numlines ? w->numlines : 1);
    scrollbar->numlines = (double) 65535.0 *box_size / (w->numlines ? w->numlines : 1);

}

int eh_NeXT_filelist (CWidget * w, XEvent * xevent, CEvent * cwevent);

CWidget *CDrawFilelist (const char *identifier, Window parent, int x, int y,
			int width, int height, int line, int column,
			struct file_entry *directentry,
			long options)
{
    struct file_entry e;
    CWidget *w;
    int x_hint;
    int lines = 0;
    unsigned int prop;

    if (!directentry) {
	memset (&e, 0, sizeof (e));
	e.options = FILELIST_LAST_ENTRY;
	directentry = &e;
    } else {
	while (!(directentry[lines].options & FILELIST_LAST_ENTRY))
	    lines++;
    }

    if (height == AUTO_HEIGHT)
	height = max (1, lines) * FONT_PIX_PER_LINE + 6;

    if (width == AUTO_WIDTH)
	width = (FONT_MEAN_WIDTH * 24 + 15);

    w = CSetupWidget (identifier, parent, x, y,
	      width, height, C_TEXTBOX_WIDGET, INPUT_KEY, COLOR_FLAT, 1);

    x_hint = x + width + WIDGET_SPACING;
    prop = (double) 65535.0 *(height / FONT_PIX_PER_LINE) / (lines ? lines : 1);
    w->vert_scrollbar = CDrawVerticalScrollbar (catstrs (identifier, ".vsc", 0), parent,
				 x_hint, y, height, AUTO_WIDTH, 0, prop);
    w->vert_scrollbar->position |= POSITION_HEIGHT;
    CSetScrollbarCallback (w->vert_scrollbar->ident, w->ident, link_scrollbar_to_NeXT_filelist);
    CGetHintPos (&x_hint, 0);

    set_hint_pos (x_hint, y + height + WIDGET_SPACING);
    w->eh = eh_NeXT_filelist;
    w->current = -1;
    w->cursor = -1;
    w->firstline = 0;
    w->numlines = lines;
    w->hook = directentry;
    return w;
}

CWidget *CRedrawFilelist (const char *identifier, struct file_entry * directentry, int preserve)
{
    CWidget *w = CIdent (identifier);

    if (w) {
	w->hook = directentry;
	if (!directentry)
	    w->numlines = 1;
	else {
	    register int i = 0;
	    while (!(directentry[i].options & FILELIST_LAST_ENTRY))
		i++;
	    w->numlines = i;
	}
	w->current = -1;
	w->cursor = -1;
	w->firstline = 0;

	free_filelist_cache (w->user);
	w->user = NULL;
	render_NeXT_filelist (w);
	if (w->user) {
	    FilelistCache *cache;
	    int box_size;
	    cache = (FilelistCache *) (w->user);
	    box_size = cache->height / cache->rowheight;
	    w->vert_scrollbar->firstline = (double) 65535.0 *w->firstline / (w->numlines ? w->numlines : 1);
	    w->vert_scrollbar->numlines = (double) 65535.0 *box_size / (w->numlines ? w->numlines : 1);
	    render_scrollbar (w->vert_scrollbar);
	}
    }
    return w;
}

void CSetFilelistPosition (const char *identifier, long current, long cursor, long firstline)
{
    CWidget *w = CIdent (identifier);

    if (w) {
	FilelistCache *cache;
	int redraw_scrollbar = 0;
	if (current >= w->numlines)
	    current = w->numlines - 1;
	if (cursor >= w->numlines)
	    cursor = w->numlines - 1;
	w->current = current;
	w->cursor = cursor;
	if (firstline >= 0 && (cache = (FilelistCache *) (w->user))) {
	    long max_fline = w->numlines - cache->height / cache->rowheight;
	    if (max_fline <= 0)
		if (firstline >= max_fline)
		    firstline = max_fline - 1;
	    /* if our window is larger then numlines : */
	    if (firstline < 0)
		firstline = 0;
	    redraw_scrollbar = (w->firstline != firstline);
	    w->firstline = firstline;
	}
	render_NeXT_filelist (w);
	if (redraw_scrollbar) {
	    w->vert_scrollbar->firstline = (double) 65535.0 *w->firstline / (w->numlines ? w->numlines : 1);
	    render_scrollbar (w->vert_scrollbar);
	}
    }
}

struct file_entry *CGetFilelistLine (CWidget * w, int line)
{
    struct file_entry *e;
    static struct file_entry r;

    memset (&r, 0, sizeof (r));
    e = (struct file_entry *) w->hook;
    if (e[line].options & FILELIST_LAST_ENTRY)
	r.options = FILELIST_LAST_ENTRY;
    else
	r = e[line];
    return &r;
}

int filelist_handle_mouse (CWidget * wdt, FilelistCache * cache, int y)
{
    int new_hilite;
    new_hilite = (int) (wdt->firstline) + y / (int) (cache->rowheight);
    if (y < 0)
	new_hilite--;
    /* we don't want to remove hilite even if user wants to go 
       beyound the beginning */
    if (new_hilite < 0)
	new_hilite = 0;

    return new_hilite;
}

int filelist_handle_keypress (CWidget * wdt, FilelistCache * cache, KeySym key)
{
    int new_cursor = wdt->cursor;

/* when text is highlighted, the cursor must be off */
    switch ((int) key) {
    case CK_Up:
	new_cursor--;
	break;
    case CK_Down:
	new_cursor++;
	break;
    case CK_Page_Up:
	new_cursor -= cache->height / cache->rowheight - 1;
	break;
    case CK_Page_Down:
	new_cursor += cache->height / cache->rowheight - 1;
	break;
    case CK_Home:
	new_cursor = 0;
	break;
    case CK_End:
	new_cursor = wdt->numlines;
	break;
    default:
	return -1;
    }
    /* we don't want to remove hilite even if user wants to go 
       beyound the beginning */
    if (new_cursor < 0)
	new_cursor = 0;
    return new_cursor;
}

int change_hilite (CWidget * w, int new_hilite)
{
    FilelistCache *cache = (FilelistCache *) (w->user);
    if (new_hilite >= w->numlines)
	new_hilite = w->numlines - 1;
    if (new_hilite != w->cursor) {
	int new_firstline = w->firstline;
	struct file_entry *directentry = (struct file_entry *) w->hook;
	if (new_hilite < w->firstline && new_hilite >= 0)
	    new_firstline = new_hilite;
	if (new_hilite >= w->firstline + w->height / cache->rowheight)
	    new_firstline = new_hilite - w->height / cache->rowheight + 1;
	if (new_firstline != w->firstline) {
	    w->cursor = new_hilite;
	    w->firstline = new_firstline;
	    w->vert_scrollbar->firstline = (double) 65535.0 *w->firstline / (w->numlines ? w->numlines : 1);
	    render_scrollbar (w->vert_scrollbar);
	}
	w->cursor = new_hilite;
	return 1;
    }
    return 0;
}

void selection_send (XSelectionRequestEvent * rq);

int eh_NeXT_filelist (CWidget * w, XEvent * xevent, CEvent * cwevent)
{
    int xevent_xbutton_y;
    int need_redraw = 0;
    FilelistCache *cache = (FilelistCache *) (w->user);
    int handled = 0;

    if ((w == NULL || w->hook == NULL || cache == NULL) &&
	xevent->type != Expose)
	return handled;

    switch (xevent->type) {
    case SelectionRequest:
/* fixme: later
   {
   int type;
   if (selection.text)  free (selection.text);
   selection.text = (unsigned char *) get_block (w, 0, 0, &type, &selection.len);
   selection_send (&(xevent->xselectionrequest));
   }
   return 1;
     */ case Expose:
	handled = 1;
	break;
    case ButtonPress:
	resolve_button (xevent, cwevent);
	need_redraw = change_hilite (w, filelist_handle_mouse (w, cache, cwevent->y));
	handled = 1;
	break;
    case ButtonRelease:
	resolve_button (xevent, cwevent);
	need_redraw = change_hilite (w, filelist_handle_mouse (w, cache, cwevent->y));
	handled = 1;
	break;
    case MotionNotify:
	resolve_button (xevent, cwevent);
	if (cwevent->state & (Button1Mask | Button2Mask)) {
	    need_redraw = change_hilite (w, filelist_handle_mouse (w, cache, cwevent->y));
	    handled = 1;
	}
	break;
    case FocusIn:
	w->options |= BUTTON_HIGHLIGHT;
	need_redraw = 1;
	handled = 1;
	break;
    case FocusOut:
	w->options &= ~BUTTON_HIGHLIGHT;
	need_redraw = 1;
	handled = 1;
	break;
    case EnterNotify:		/* we simply let our parent to process it */
    case LeaveNotify:
	cwevent->ident = w->ident;
	break;
    case KeyPress:
	{
	    int move_to = filelist_handle_keypress (w, cache, cwevent->command);

	    if (move_to >= 0) {
		need_redraw = change_hilite (w, move_to);
		handled = 1;
	    } else {
		switch (cwevent->command) {
		case CK_Left:
		case CK_Right:
		case CK_Home:
		case CK_End:
		case CK_Enter:
		case CK_BackSpace:
		    cwevent->ident = w->ident;
		    return 1;
		default:
		    if (cwevent->insert > 0) {
			cwevent->ident = w->ident;
			return 1;
		    }
		}
	    }
	}
	break;

    default:
	return handled;
    }

    if (need_redraw || cwevent->double_click)
	cwevent->ident = w->ident;

    if ((xevent->type == Expose && !xevent->xexpose.count) || need_redraw)
	render_NeXT_filelist (w);

    return handled;
}
