/* scrollbar_next.c scrollbar widget with NeXT ish look and feel
   Copyright (C) 1999 Sasha Vasko
   based on :
   scrollbar.c - scrollbar widget
   Copyright (C) 1997 Paul Sheer

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.
 */


#include <config.h>
#include <stdio.h>
#include <my_string.h>
#include <stdlib.h>
#include <stdarg.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xresource.h>
#include <X11/Xatom.h>
#include "lkeysym.h"

#include "stringtools.h"
#include "app_glob.c"
#include "edit.h"
#include "editcmddef.h"

#include "coolwidget.h"
#include "coollocal.h"
#include "mousemark.h"

#include "mad.h"

#define PURENEXT
/*****************************************/
/* start NeXT scrollbar specific fetures */
/* could be anything from 13 to 19  but the true NeXT is 17 */
#define SB_WIDTH		17
/* this will define somemore parameters for shaping NeXTish scrollbars */
/* NEXT_SCROLL_CLEAN if defined removes shades of grey from buttons */
#undef NEXT_SCROLL_CLEAN
#define NEXT_SCROLL_SQUARE_ARROWS
#define SB_BORDER_WIDTH 1
/* this makes buttons thinner then scrollbar's base ( if more then 0 ) */
#define SIDE_STEP_WIDTH 1
/*  end NeXT scrollbar specific fetures  */
/*****************************************/

/*****************************************/
/* scrollbarr configuration stuff        */

#define Xdisplay CDisplay
#define Xroot    CRoot
#define Xdepth   CDepth

static char *SCROLLER_DIMPLE[] =
{
    ".%###.",
    "%#%%%%",
    "#%%...",
    "#%..  ",
    "#%.   ",
    ".%.  ."
};

#define SCROLLER_DIMPLE_WIDTH   6
#define SCROLLER_DIMPLE_HEIGHT  6

static char *SCROLLER_ARROW[4][13] =
{
    {".............",
     ".............",
     "......%......",
     "......#......",
     ".....%#%.....",
     ".....###.....",
     "....%###%....",
     "....#####....",
     "...%#####%...",
     "...#######...",
     "..%#######%..",
     ".............",
     "............."
    },
    {".............",
     ".............",
     "..%#######%..",
     "...#######...",
     "...%#####%...",
     "....#####....",
     "....%###%....",
     ".....###.....",
     ".....%#%.....",
     "......#......",
     "......%......",
     ".............",
     "............."
    },
    {"             ",
     "             ",
     "      %      ",
     "      %      ",
     "     %%%     ",
     "     %%%     ",
     "    %%%%%    ",
     "    %%%%%    ",
     "   %%%%%%%   ",
     "   %%%%%%%   ",
     "  %%%%%%%%%  ",
     "             ",
     "             "
    },
    {"             ",
     "             ",
     "  %%%%%%%%%  ",
     "   %%%%%%%   ",
     "   %%%%%%%   ",
     "    %%%%%    ",
     "    %%%%%    ",
     "     %%%     ",
     "     %%%     ",
     "      %      ",
     "      %      ",
     "             ",
     "             "
    }};

#define ARROW_SOURCE_WIDTH   13
#define ARROW_SOURCE_HEIGHT  13

typedef struct {
    Pixmap icon;
    Pixmap icon_mask;
    int origin_x, origin_y, width, height;
} Icon;

typedef struct {
    unsigned arrow_width, arrow_height;
    int bValid;
    Icon dimple;
    Icon Arrows[8];
    Pixmap stipple;
    /* this stuff is related to cache */
    XRectangle thumb;
    XRectangle buttons[2];
    unsigned int button_arrow[2];
    unsigned int width, height;
    int work_length;
    Pixmap cache;
} ScrollIcons;

#define UP_ARROW	(next_icons->Arrows[0])
#define UP_ARROW_HI	(next_icons->Arrows[1])
#define DOWN_ARROW	(next_icons->Arrows[2])
#define DOWN_ARROW_HI	(next_icons->Arrows[3])

#define ARROW_WIDTH   (next_icons->arrow_width)
#define ARROW_HEIGHT  (next_icons->arrow_height)

#define BEVEL_HI_WIDTH 1
#define BEVEL_LO_WIDTH 2
#define BEVEL_SIZE (BEVEL_HI_WIDTH+BEVEL_LO_WIDTH)

#define SB_BUTTON_HEIGHT (BEVEL_SIZE+ARROW_HEIGHT)
#define SB_BUTTONS_HEIGHT ((SB_BUTTON_HEIGHT<<1)-SIDE_STEP_WIDTH)

#ifndef SB_BORDER_WIDTH
#define SB_BORDER_WIDTH 1
#endif
#define SB_BORDER_SIZE  (SB_BORDER_WIDTH<<1)

#ifndef SIDE_STEP_WIDTH
#define SIDE_STEP_WIDTH 0
#endif

#define SB_MIN_THUMB_SIZE  (SCROLLER_DIMPLE_WIDTH+2+BEVEL_SIZE)

/* end NeXT unconfigurable stuff */
/*********************************/

#define stp_width 8
#define stp_height 8
static unsigned char stp_bits[] =
{
    0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa};

#undef  TRANSPARENT
#define IS_TRANSP_SCROLL 0


typedef struct {
    GC blackGC;
    GC whiteGC;
    GC darkGC;
    GC maskGC;
    GC maskGC_0;
} IconGC;

static void init_scroll_size (ScrollIcons * next_icons)
{
    ARROW_WIDTH = (SB_WIDTH - BEVEL_SIZE - SB_BORDER_SIZE - SIDE_STEP_WIDTH);
    ARROW_WIDTH = min (ARROW_WIDTH, ARROW_SOURCE_WIDTH);

#ifdef NEXT_SCROLL_SQUARE_ARROWS
    ARROW_HEIGHT = ARROW_WIDTH;
#else
    ARROW_HEIGHT = ARROW_SOURCE_HEIGHT;
#endif
    next_icons->bValid = 1;
}

/* PROTO */
unsigned GetScrollArrowsHeight (ScrollIcons * next_icons)
{
    if (!next_icons->bValid)
	init_scroll_size (next_icons);
    return (SB_BUTTONS_HEIGHT);
}

static void CheckIconGC (IconGC * igc, Pixmap icon, Pixmap icon_mask)
{
    XGCValues values;
    unsigned long valuemask = GCForeground | GCGraphicsExposures;

    values.graphics_exposures = False;

    if (igc == NULL)
	return;
    if (igc->maskGC == None) {
	values.foreground = 1;
	igc->maskGC = XCreateGC (Xdisplay, icon_mask, valuemask, &values);
    }
    if (igc->maskGC_0 == None) {
	values.foreground = 0;
	igc->maskGC_0 = XCreateGC (Xdisplay, icon_mask, valuemask, &values);
    }
    if (igc->whiteGC == None) {
	values.foreground = COLOR_WHITE;
	igc->whiteGC = XCreateGC (Xdisplay, icon, valuemask, &values);
    }
    if (igc->darkGC == None) {
	values.foreground = COLOR_DARK;
	igc->darkGC = XCreateGC (Xdisplay, icon, valuemask, &values);
    }
    if (igc->blackGC == None) {
	values.foreground = COLOR_BLACK;
	igc->blackGC = XCreateGC (Xdisplay, icon, valuemask, &values);
    }
}

static void FreeIconGC (IconGC * igc)
{
    if (igc) {
	if (igc->maskGC != None) {
	    XFreeGC (Xdisplay, igc->maskGC);
	    igc->maskGC = None;
	}
	if (igc->maskGC_0 != None) {
	    XFreeGC (Xdisplay, igc->maskGC_0);
	    igc->maskGC_0 = None;
	}
	if (igc->whiteGC == None) {
	    XFreeGC (Xdisplay, igc->whiteGC);
	    igc->whiteGC = None;
	}
	if (igc->darkGC == None) {
	    XFreeGC (Xdisplay, igc->darkGC);
	    igc->darkGC = None;
	}
	if (igc->blackGC != None) {
	    XFreeGC (Xdisplay, igc->blackGC);
	    igc->blackGC = None;
	}
    }
}

static void renderIcon (Window win, char **data, Icon * pIcon, IconGC * igc, Bool rotate)
{
    Pixmap d, mask;
    register int i, k;
    int x, y, max_x, max_y;
    GC maskgc, paintgc;
    char pixel;

    d = XCreatePixmap (Xdisplay, win, pIcon->width, pIcon->height, Xdepth);
    mask = XCreatePixmap (Xdisplay, win, pIcon->width, pIcon->height, 1);

    if (rotate) {
	max_x = pIcon->height;
	max_y = pIcon->width;
    } else {
	max_x = pIcon->width;
	max_y = pIcon->height;
    }

    CheckIconGC (igc, d, mask);
    y = pIcon->origin_y;

    for (i = 0; i < max_y; y++, i++) {
	x = pIcon->origin_x;
	for (k = 0; k < max_x; k++, x++) {
	    maskgc = igc->maskGC;
	    if (rotate)
		pixel = data[x][y];
	    else
		pixel = data[y][x];
	    switch (pixel) {
	    case ' ':
	    case 'w':
		paintgc = igc->whiteGC;
		break;
	    case '%':
	    case 'd':
		paintgc = igc->darkGC;
		break;
	    case '#':
	    case 'b':
		paintgc = igc->blackGC;
		break;
	    case '.':
	    case 'l':
	    default:
		paintgc = CGC;
		maskgc = igc->maskGC_0;
		break;
	    }
	    XDrawPoint (Xdisplay, d, paintgc, k, i);
	    XDrawPoint (Xdisplay, mask, maskgc, k, i);
	}
    }

    pIcon->icon = d;
    pIcon->icon_mask = mask;
}

static ScrollIcons *
 init_next_icons (Window win)
{
    unsigned arrow_x_offset, arrow_y_offset;
    IconGC icongc =
    {None, None, None, None};
    int i;
    ScrollIcons *next_icons = calloc (1, sizeof (ScrollIcons));

    if (next_icons == NULL)
	return next_icons;

    CSetColor (COLOR_FLAT);

    next_icons->dimple.width = SCROLLER_DIMPLE_WIDTH;
    next_icons->dimple.height = SCROLLER_DIMPLE_WIDTH;
    renderIcon (win, SCROLLER_DIMPLE, &(next_icons->dimple), &icongc, False);

    init_scroll_size (next_icons);
    arrow_x_offset = (ARROW_SOURCE_WIDTH - ARROW_WIDTH) >> 1;
#ifdef NEXT_SCROLL_SQUARE_ARROWS
    arrow_y_offset = arrow_x_offset;
#else
    arrow_y_offset = 0;		/* not implemented yet */
#endif

    for (i = 0; i < 4; i++) {
	next_icons->Arrows[i].origin_x = arrow_x_offset;
	next_icons->Arrows[i].origin_y = arrow_y_offset;
	next_icons->Arrows[i].width = ARROW_WIDTH;
	next_icons->Arrows[i].height = ARROW_HEIGHT;
	renderIcon (win, SCROLLER_ARROW[i], &(next_icons->Arrows[i]), &icongc, False);
    }
    for (; i < 8; i++) {
	next_icons->Arrows[i].origin_x = arrow_y_offset;
	next_icons->Arrows[i].origin_y = arrow_x_offset;
	next_icons->Arrows[i].width = ARROW_HEIGHT;
	next_icons->Arrows[i].height = ARROW_WIDTH;
	renderIcon (win, SCROLLER_ARROW[i - 4], &(next_icons->Arrows[i]), &icongc, True);
    }

    FreeIconGC (&icongc);

    next_icons->stipple = XCreatePixmap (Xdisplay, win, 2, 2, Xdepth);
    XDrawPoint (Xdisplay, next_icons->stipple, CGC, 0, 0);
    XDrawPoint (Xdisplay, next_icons->stipple, CGC, 1, 1);

    CSetColor (COLOR_DARK);
    XDrawPoint (Xdisplay, next_icons->stipple, CGC, 0, 1);
    XDrawPoint (Xdisplay, next_icons->stipple, CGC, 1, 0);

    return next_icons;
}

void free_next_icons (void *ptr)
{
    ScrollIcons *next_icons = (ScrollIcons *) ptr;
    if (next_icons) {
	register int i;
	if (next_icons->dimple.icon) {
	    XFreePixmap (Xdisplay, next_icons->dimple.icon);
	    next_icons->dimple.icon = None;
	}
	if (next_icons->stipple) {
	    XFreePixmap (Xdisplay, next_icons->stipple);
	    next_icons->stipple = None;
	}
	for (i = 0; i < 4; i++)
	    if (next_icons->Arrows[i].icon) {
		XFreePixmap (Xdisplay, next_icons->Arrows[i].icon);
		next_icons->Arrows[i].icon = None;
	    }
	if (next_icons->cache) {
	    XFreePixmap (Xdisplay, next_icons->cache);
	    next_icons->cache = None;
	}
    }
}

static void render_next_icon (Drawable d, Icon * i, int x, int y)
{
#ifdef TRANSPARENT
    if (IS_TRANSP_SCROLL) {
	XSetClipMask (Xdisplay, CGC, i->icon_mask);
	XSetClipOrigin (Xdisplay, CGC, x, y);
    }
#endif
    XCopyArea (Xdisplay, i->icon, d, CGC, 0, 0,
	       i->width, i->height, x, y);

#ifdef TRANSPARENT
    if (IS_TRANSP_SCROLL)
	XSetClipMask (Xdisplay, CGC, None);
#endif
}


/* Draw bezel & arrows */
static void render_next_button (Drawable d, XRectangle * rec, Icon * icon, Bool pressed)
{
    int x2 = rec->x + rec->width - 1, y2 = rec->y + rec->height - 1;
    CSetColor ((pressed) ? COLOR_BLACK : COLOR_WHITE);

    CLine (d, rec->x, rec->y, rec->x, y2);
    CLine (d, rec->x, rec->y, x2, rec->y);

    CSetColor ((pressed) ? COLOR_WHITE : COLOR_BLACK);
    CLine (d, rec->x, y2, x2, y2);
    CLine (d, x2, y2, x2, rec->y);

    if (!pressed)
	CSetColor (COLOR_FLAT);

    CRectangle (d, rec->x + 1, rec->y + 1, rec->width - 3, rec->height - 3);

    render_next_icon (d, icon, rec->x + ((rec->width - icon->width) / 2),
		      rec->y + ((rec->height - icon->height) / 2));
}

Bool check_cache (ScrollIcons * next_icons, Window win, int width, int height, int pos, int prop, int flags)
{
    unsigned int l, wl, thumb_start, thumb_length;
    unsigned int button_arrow[2];
    int cache_valid = 0;

    if (width > height) {
	if (height > SB_WIDTH + SB_BORDER_SIZE)		/* vertical */
	    height = SB_WIDTH + SB_BORDER_SIZE;
	l = width;
    } else {
	if (width > SB_WIDTH + SB_BORDER_SIZE)
	    width = SB_WIDTH + SB_BORDER_SIZE;
	l = height;
    }
    wl = l - SB_BORDER_SIZE - SB_BUTTONS_HEIGHT - SIDE_STEP_WIDTH;

    button_arrow[0] = 0;
    button_arrow[1] = 1;

    if (prop == 65535 || prop == 0) {
	thumb_length = 0;
	thumb_start = 0;
    } else {
	thumb_length = (double) (wl * prop) / 65535.0;
	if (thumb_length < SB_MIN_THUMB_SIZE) {
	    wl -= SB_MIN_THUMB_SIZE - thumb_length;
	    thumb_length = SB_MIN_THUMB_SIZE;
	} else if (thumb_length > wl)
	    thumb_length = wl;

	if (pos >= 65535 - prop - 1)
	    thumb_start = l - thumb_length - SB_BUTTONS_HEIGHT;
	else
	    thumb_start = ((double) (wl * pos) / 65535.0) + SB_BORDER_WIDTH;

	if (thumb_start - SB_BORDER_WIDTH > wl - thumb_length)
	    thumb_start = SB_BORDER_WIDTH + wl - thumb_length;
	if (thumb_start < SB_BORDER_WIDTH)
	    thumb_start = SB_BORDER_WIDTH;

	next_icons->work_length = wl;

	if (!(flags & 0x20)) {
	    if ((flags & 0x0f) == 2)
		button_arrow[0] = 2;
	    if ((flags & 0x0f) == 5)
		button_arrow[1] = 3;
	}
    }
    if (l != height) {
	button_arrow[0] += 4;
	button_arrow[1] += 4;
    }
    if (next_icons->cache) {
	if (width == next_icons->width && height == next_icons->height) {
	    if (button_arrow[0] == next_icons->button_arrow[0] &&
		button_arrow[1] == next_icons->button_arrow[1]) {
		if (l == height) {
		    if (next_icons->thumb.height == thumb_length &&
			next_icons->thumb.y == thumb_start)
			return True;
		} else if (next_icons->thumb.width == thumb_length &&
			   next_icons->thumb.x == thumb_start)
		    return True;
	    }
	} else {
	    XFreePixmap (Xdisplay, next_icons->cache);
	    next_icons->cache = None;
	}
    }
    next_icons->button_arrow[0] = button_arrow[0];
    next_icons->button_arrow[1] = button_arrow[1];

    if (l == height) {
	next_icons->thumb.height = thumb_length;
	next_icons->thumb.y = thumb_start;
	next_icons->thumb.width = SB_WIDTH;
	next_icons->thumb.x = SB_BORDER_WIDTH;
    } else {
	next_icons->thumb.width = thumb_length;
	next_icons->thumb.x = thumb_start;
	next_icons->thumb.height = SB_WIDTH;
	next_icons->thumb.y = SB_BORDER_WIDTH;
    }
    if (next_icons->cache == None) {	/* create double buffer */
	next_icons->cache = XCreatePixmap (Xdisplay, win, width, height, Xdepth);
	next_icons->width = width;
	next_icons->height = height;
	if (l == height) {
	    next_icons->buttons[0].x = next_icons->buttons[1].x = SB_BORDER_WIDTH;
	    next_icons->buttons[0].width = next_icons->buttons[1].width = SB_WIDTH;
	    next_icons->buttons[0].height = next_icons->buttons[1].height = SB_BUTTON_HEIGHT;
	    next_icons->buttons[1].y = l - next_icons->buttons[1].height - SIDE_STEP_WIDTH;
	    next_icons->buttons[0].y = next_icons->buttons[1].y - next_icons->buttons[0].height;
	} else {
	    next_icons->buttons[0].y = next_icons->buttons[1].y = SB_BORDER_WIDTH;
	    next_icons->buttons[0].height = next_icons->buttons[1].height = SB_WIDTH;
	    next_icons->buttons[0].width = next_icons->buttons[1].width = SB_BUTTON_HEIGHT;
	    next_icons->buttons[1].x = l - next_icons->buttons[1].width - SIDE_STEP_WIDTH;
	    next_icons->buttons[0].x = next_icons->buttons[1].x - next_icons->buttons[0].width;
	}
    }
    return False;
}

void scrollbar_fill_back (ScrollIcons * next_icons)
{
    int x2, y2;
    register int i;

    /* draw the background */
    XSetTile (Xdisplay, CGC, next_icons->stipple);
    XSetFillStyle (Xdisplay, CGC, FillTiled);
    XSetTSOrigin (Xdisplay, CGC, 0, 0);

    CSetColor (COLOR_WHITE);
    CRectangle (next_icons->cache,
		SB_BORDER_WIDTH, SB_BORDER_WIDTH,
		(next_icons->width) - SB_BORDER_SIZE,
		(next_icons->height) - SB_BORDER_SIZE);

    XSetTile (Xdisplay, CGC, None);
    XSetFillStyle (Xdisplay, CGC, FillSolid);

    x2 = next_icons->width - 1;
    y2 = next_icons->height - 1;

    CSetColor (COLOR_WHITE);
    for (i = 0; i < SB_BORDER_WIDTH; i++) {
	CLine (next_icons->cache, i, y2 - i, x2 - i, y2 - i);
	CLine (next_icons->cache, x2 - i, y2 - i, x2 - i, i);
    }
    CSetColor (COLOR_BLACK);
    for (i = 0; i < SB_BORDER_WIDTH; i++) {
	CLine (next_icons->cache, i, y2 - i, i, i);
	CLine (next_icons->cache, i, i, x2 - i, i);
    }


}

int render_next_scrollbar (ScrollIcons * next_icons, Window win, int x, int y, int width, int height, int pos, int prop, int flags)
{
    int cache_valid;

    cache_valid = check_cache (next_icons, win, width, height, pos, prop, flags);
/*fprintf( stderr, "%dx%d,pos: %d, prop: %d ->%s, whole: %dx%d, thumb: %dx%d%+d%+d\n",
   width, height, pos, prop, 
   (cache_valid)?"Cached":"Redrawing",
   next_icons->width, next_icons->height,
   next_icons->thumb.width, next_icons->thumb.height,
   next_icons->thumb.x, next_icons->thumb.y ); 
 */ if (!cache_valid) {
	int i;
	scrollbar_fill_back (next_icons);
	if (next_icons->thumb.width > 0 &&
	    next_icons->thumb.height > 0) {
	    render_next_button (next_icons->cache,
				&(next_icons->thumb),
				&(next_icons->dimple),
				False);
	    for (i = 0; i < 2; i++) {
		render_next_button (next_icons->cache,
				    &(next_icons->buttons[i]),
		      &(next_icons->Arrows[next_icons->button_arrow[i]]),
				    (next_icons->button_arrow[i] & 0x02));
	    }
	}
    }
    XCopyArea (Xdisplay, next_icons->cache, win, CGC, 0, 0, next_icons->width, next_icons->height, 0, 0);
    return 1;
}



/************************************************************************/
/* old cooledit interface */

CWidget *CDrawVerticalScrollbar (const char *identifier, Window parent, int x, int y,
				 int length, int width, int pos, int prop)
{
    CWidget *w;
    if (width == AUTO_WIDTH || width == AUTO_HEIGHT)
	width = SB_WIDTH + SB_BORDER_SIZE;
    w = CSetupWidget (identifier, parent, x, y,
	width, length, C_VERTSCROLL_WIDGET, INPUT_MOTION, COLOR_FLAT, 0);
    w->firstline = pos;
    w->numlines = prop;
    set_hint_pos (x + width + WIDGET_SPACING, y + length + WIDGET_SPACING);

    w->user = init_next_icons (w->winid);
    w->free_user = free_next_icons;

    return w;
}

CWidget *CDrawHorizontalScrollbar (const char *identifier, Window parent, int x, int y,
				int length, int width, int pos, int prop)
{
    CWidget *w;
    if (width == AUTO_WIDTH || width == AUTO_HEIGHT)
	width = SB_WIDTH + SB_BORDER_SIZE;
    w = CSetupWidget (identifier, parent, x, y,
		      length, width, C_HORISCROLL_WIDGET,
		      ExposureMask | ButtonPressMask |
	       ButtonReleaseMask | ButtonMotionMask | PointerMotionMask |
		      EnterWindowMask | LeaveWindowMask, COLOR_FLAT, 0);
    w->firstline = pos;
    w->numlines = prop;
    set_hint_pos (x + length + WIDGET_SPACING, y + width + WIDGET_SPACING);

    w->user = init_next_icons (w->winid);
    w->free_user = free_next_icons;

    return w;
}

void CSetScrollbarCallback (const char *scrollbar, const char *wdt,
			    void (*link_to) (CWidget *,
				     CWidget *, XEvent *, CEvent *, int))
{
    CWidget *s, *w;
    s = CIdent (scrollbar);
    w = CIdent (wdt);
    if (!s || !w)
	return;
    s->vert_scrollbar = w;
    s->scroll_bar_link = link_to;
}

void render_scrollbar (CWidget * wdt)
{
    int pos, prop;
    int flags = wdt->options;
    if (!wdt)
	return;

    pos = wdt->firstline;
    prop = wdt->numlines;

    if (prop < 0)
	prop = 0;
    if (pos < 0)
	pos = 0;
    if (pos > 65535)
	pos = 65535;
    if (pos + prop >= 65535)
	prop = 65535 - pos;

    render_next_scrollbar ((ScrollIcons *) wdt->user, wdt->winid,
			   wdt->x, wdt->y, wdt->width, wdt->height,
			   pos, prop, flags);

    if (wdt->scroll_bar_extra_render)
	(*wdt->scroll_bar_extra_render) (wdt);
}

/*
   Which scrollbar button was pressed: 3 is the middle button ?
 */
int which_scrollbar_button (int bx, int by, CWidget * wdt)
{
    int pos;
    int prop;
    ScrollIcons *next_icons = (ScrollIcons *) wdt->user;
/*    
   fprintf( stderr, "button: (%d,%d) thumb: (%d,%d)\n",bx, by, next_icons->thumb.x, next_icons->thumb.y  );    
 */
    if (wdt->kind == C_VERTSCROLL_WIDGET) {
	if (by > 0 && by < next_icons->thumb.y)
	    return 1;
	if (by >= next_icons->thumb.y &&
	    by <= next_icons->thumb.y + next_icons->thumb.height)
	    return 3;
	if (by > next_icons->thumb.y + next_icons->thumb.height &&
	    by < next_icons->buttons[0].y)
	    return 4;
	if (by >= next_icons->buttons[0].y &&
	  by <= next_icons->buttons[0].y + next_icons->buttons[0].height)
	    return 2;
	if (by >= next_icons->buttons[1].y &&
	  by <= next_icons->buttons[1].y + next_icons->buttons[1].height)
	    return 5;
    } else {
	if (bx > 0 && bx < next_icons->thumb.x)
	    return 1;
	if (bx >= next_icons->thumb.x &&
	    bx <= next_icons->thumb.x + next_icons->thumb.width)
	    return 3;
	if (bx > next_icons->thumb.x + next_icons->thumb.width &&
	    bx < next_icons->buttons[0].x)
	    return 4;
	if (bx >= next_icons->buttons[0].x &&
	    bx <= next_icons->buttons[0].x + next_icons->buttons[0].width)
	    return 2;
	if (bx >= next_icons->buttons[1].x &&
	    bx <= next_icons->buttons[1].x + next_icons->buttons[1].width)
	    return 5;

    }
    return 0;
}

int eh_scrollbar (CWidget * w, XEvent * xevent, CEvent * cwevent)
{
    static int buttonypos, y, offset, whichscrbutton = 0;	/* which of the five scroll bar buttons was pressed */
    int xevent_xbutton_y, length, width, thumb_pos;
    ScrollIcons *next_icons = (ScrollIcons *) w->user;

    if (w->kind == C_VERTSCROLL_WIDGET) {
	xevent_xbutton_y = xevent->xbutton.y;
	length = w->height - SB_BUTTONS_HEIGHT - SB_BORDER_SIZE - next_icons->thumb.height;
	width = w->width;
	thumb_pos = next_icons->thumb.y;
    } else {
	xevent_xbutton_y = xevent->xbutton.x;
	length = w->width - SB_BUTTONS_HEIGHT - SB_BORDER_SIZE - next_icons->thumb.width;
	width = w->height;
	thumb_pos = next_icons->thumb.x;
    }

    if (next_icons->thumb.width == 0 || next_icons->thumb.height == 0) {
	if (xevent->type == Expose && !xevent->xexpose.count)
	    render_scrollbar (w);
	return 0;
    }
    switch (xevent->type) {
/*    case LeaveNotify:
 */
    case Expose:
	w->options = whichscrbutton = 0;
	break;
    case ButtonRepeat:
	resolve_button (xevent, cwevent);
	if (whichscrbutton != 3 &&
	    (cwevent->button == Button1 || cwevent->button == Button2)) {
	    int b;
	    b = which_scrollbar_button (cwevent->x, cwevent->y, w);
	    if (b != 3 && b > 0) {
		y = w->firstline;
		buttonypos = xevent_xbutton_y;
		w->options = whichscrbutton = b;
		cwevent->ident = w->ident;
		xevent->type = cwevent->type = ButtonPress;
	    }
	}
	break;
    case ButtonPress:
	resolve_button (xevent, cwevent);
	if (cwevent->button == Button1 || cwevent->button == Button2) {
	    buttonypos = xevent_xbutton_y;
	    y = w->firstline;
	    w->options = whichscrbutton = which_scrollbar_button (cwevent->x, cwevent->y, w);
	    if (whichscrbutton == 3) {
		offset = thumb_pos - buttonypos + SB_BORDER_WIDTH;
	    } else
		offset = 0;
	    cwevent->ident = w->ident;
	    w->search_start = w->firstline;
	    w->search_len = w->numlines;
	}
	break;
    case ButtonRelease:
	resolve_button (xevent, cwevent);
	if (whichscrbutton == 3) {
	    w->options = 0x20 + whichscrbutton;
	    y = (double) (xevent_xbutton_y + offset) * (double) 65535.0 / next_icons->work_length;
	    w->firstline = y;
	    buttonypos = xevent_xbutton_y;
	}
	break;
    case MotionNotify:
	if (whichscrbutton == 0)
	    return 0;
	resolve_button (xevent, cwevent);
	if (cwevent->state & (Button1Mask | Button2Mask)) {
	    w->options = whichscrbutton;
	    if (whichscrbutton == 3) {
		y = (double) (xevent_xbutton_y + offset) * (double) 65535.0 / (next_icons->work_length);
		w->firstline = y;
		buttonypos = xevent_xbutton_y;
	    }
	}
	break;
    default:
	return 0;
    }

    if (w->firstline > 65535)
	w->firstline = 65535;
    if (cwevent->state & (Button1Mask | Button2Mask) || cwevent->type == ButtonPress || cwevent->type == ButtonRelease)
	if (w->scroll_bar_link && w->vert_scrollbar)
	    (*w->scroll_bar_link) (w, w->vert_scrollbar, xevent, cwevent, whichscrbutton);

    if (cwevent->type == ButtonRelease)
	w->options = whichscrbutton = 0;
    if (xevent->type != Expose || !xevent->xexpose.count)
	render_scrollbar (w);

    return 0;
}

int inbounds (int x, int y, int x1, int y1, int x2, int y2)
{
    if (x >= x1 && x <= x2 && y >= y1 && y <= y2)
	return 1;
    else
	return 0;
}
